# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .emailaddress import EmailAddress
from .genericpostaladdress import GenericPostalAddress
from .iso3166 import ISO3166Alpha2
from .list_ import List
from .phonenumber import Phonenumber
from .romanizedname import RomanizedName
from .symbolicname import SymbolicName


__all__: list[str] = [
    'EmailAddress',
    'GenericPostalAddress',
    'ISO3166Alpha2',
    'List',
    'Phonenumber',
    'RomanizedName',
    'SymbolicName',
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Any
from typing import TypeVar

import pydantic


T = TypeVar('T')


class IMeta(pydantic.BaseModel):

    @classmethod
    def clone(
        cls: T,
        *,
        named: bool = False,
        namespaced: bool = False,
        annotations: bool = False,
        labels: bool = False,
        pristine: bool = False,
        **kwargs: Any
    ) -> T:
        exclude: set[str] = set()
        if not annotations:
            exclude.add('annotations')
        if not namespaced:
            exclude.add('namespace')
        if not pristine:
            exclude.add('generate_name')
        if not labels:
            exclude.add('labels')
        new_class = type(cls.__name__, (IMeta,), { # type: ignore
            '__annotations__': {
                k: v.annotation # type: ignore
                for k, v in cls.__fields__.items() #type: ignore
                if k not in exclude
            },
            **{
                k: v.field_info # type: ignore
                for k, v in cls.__fields__.items() # type: ignore
                if k not in exclude
            },
            'Config': type('Config', (cls.Config,), { # type: ignore
                'exclude': exclude
            })
        })
        new_class.__fields__['name'].required = named
        return new_class

    def must_generate_name(self) -> bool:
        return False

    class Config:
        allow_population_by_field_name: bool = True


class BaseMeta(IMeta):
    name: str | None = pydantic.Field(
        default=None,
        title="Name",
        description=(
            "Name must be unique within a namespace. Is required when creating "
            "resources, although some resources may allow a client to request "
            "the generation of an appropriate name automatically. Name is "
            "primarily intended for creation idempotence and configuration "
            "definition. Cannot be updated.\n\n"
            "If the resource is not identified by symbolic names, but an "
            "auto-generated identifier instead, this field contains the "
            "same value as the `uid` field."
        )
    )

    namespace: str | None = pydantic.Field(
        default=None,
        title="Namespace",
        description=(
            "Namespace defines the space within which each name must be "
            "unique. An empty namespace is equivalent to the `default` "
            "namespace, but `default` is the canonical representation. "
            "Not all objects are required to be scoped to a namespace "
            "- the value of this field for those objects will be empty."
            "Not all servers may implement the concept of namespaces."
        )
    )

    labels: dict[str, str] = pydantic.Field(
        default={},
        title='Labels',
        description=(
            "Map of string keys and values that can be used to "
            "organize and categorize (scope and select) objects."
        )
    )

    annotations: dict[str, Any] = pydantic.Field(
        default={},
        title='Annotations',
        description=(
            "Annotations is an unstructured key value map stored "
            "with a resource that may be set by external tools "
            "to store and retrieve arbitrary metadata. They are "
            "not queryable and should be preserved when modifying "
            "objects."
        )
    )

    generate_name: bool | None = pydantic.Field(
        default=None,
        title="Generate name",
        alias='generateName',
        description=(
            "The `generateName` field is an optional prefix, used by the server, "
            "to generate a unique name ONLY IF the `name` field has not been "
            "provided. If this field is used, the `name` returned to the client "
            "will be different than the name passed. This value will also be "
            "combined with a unique suffix. The provided value has the same "
            "validation rules as the Name field, and may be truncated by the "
            "length of the suffix required to make the value unique on the server."
            "\n\n"
            "If this field is specified and the generated name exists, the "
            "server will return a 409."
        )
    )

    def must_generate_name(self) -> bool:
        return getattr(self, 'generate_name', False)

    class Config(IMeta.Config):
        pass


class ObjectMeta(BaseMeta):
    generation: int = pydantic.Field(
        default=0,
        title='Generation',
        description=(
            "A sequence number representing a specific generation "
            "of the desired state. Populated by the system. Read-only."
        )
    )

    creation_timestamp: datetime.datetime | None = pydantic.Field(
        default=None,
        title="Created",
        alias='creationTimestamp',
        description=(
            "The `creationTimestamp` field is a timestamp representing "
            "the server time when this object was created. It is not "
            "guaranteed to be set in happens-before order across "
            "separate operations. Clients may not set this value and "
            "is ignored when provided."
            "It is represented in RFC3339 form and is in UTC."
        )
    )

    resource_version: str = pydantic.Field(
        default='pristine',
        title='Resource version',
        alias='resourceVersion',
        description=(
            "An opaque value that represents the internal version of "
            "this object that can be used by clients to determine when "
            "objects have changed. May be used for optimistic concurrency, "
            "change detection, and the watch operation on a resource or "
            "set of resources. Clients must treat these values as opaque "
            "and passed unmodified back to the server. They may only be "
            "valid for a particular resource or set of resources.\n\n"
            "Populated by the system. Read-only. Value must be treated "
            "as opaque by clients."
        )
    )

    uid: int | str | None = pydantic.Field(
        default=None,
        title="UID",
        description=(
            "UID is the unique in time and space value for this object. It "
            "is typically generated by the server on successful creation of "
            "a resource and is not allowed to change on PUT operations.\n\n"
            "Depending on the implementation, the `uid` field may be either "
            "an integer or a string, but clients should treat this value "
            "as opaque."
        )
    )

    links: dict[str, pydantic.AnyUrl] = pydantic.Field(
        default={},
        title='Links',
        description=(
            "A mapping of operations to links that may be used by clients "
            "to traverse the REStful API and discover additional "
            "methods on resources."
        )
    )

    class Config(BaseMeta.Config):
        allow_population_by_field_name: bool = True
        fields: dict[str, dict[str, Any]] = {
            'generate_name': {'exclude': True}
        }


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .iso3166 import ISO3166Alpha2


class GenericPostalAddress(pydantic.BaseModel):
    """A generic postal address. It is assumed that further validation
    is done upstream.
    """
    addressee: str = pydantic.Field(
        default=...,
        title='Address',
        description=(
            "The addressee line of the address. If the receiver is a person, "
            "this field contains its name. For organizations, the receiver may "
            "be further specified using the `contact_name` field."
        )
    )

    contact_name: str = pydantic.Field(
        default='',
        title='Contact name',
        alias='contactName',
        description=(
            "The name of the receiver. Use this field when sending, for example, "
            "to an organization with a specific person."
        )
    )

    address1: str = pydantic.Field(
        default=...,
        title='Address #1',
        description=(
            "First line of the address, usually the street/thoroughfare "
            "name. May also be a mailbox name/number."
        )
    )

    address2: str = pydantic.Field(
        default="",
        title='Address #2',
        description=(
            "Second line of the address."
        )
    )

    postal_code: str = pydantic.Field(
        default=...,
        alias='postalCode',
        title='Postal code'
    )

    city: str = pydantic.Field(
        default=...,
        title='City'
    )

    region: str = pydantic.Field(
        default='',
        title='Region',
        description=(
            "The first-order administrative division in which the address "
            "is located, such as a state in the United States."
        )
    )

    country: ISO3166Alpha2 = pydantic.Field(
        default=...,
        title='Country',
        description="The country represented as an ISO 3166 country code."
    )

    class Config:
        allow_population_by_field_name: bool = True


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from types import EllipsisType
from typing import Any
from typing import Literal
from typing import TypeVar

import inflect
import pydantic

from .objectmeta import BaseMeta
from .objectmeta import ObjectMeta


T = TypeVar('T')
engine = inflect.engine()


class Config:
    named: bool = False
    namespaced: bool = False
    annotations: bool = False
    labels: bool = False
    pluralname: str

    @classmethod
    def dict(cls) -> dict[str, Any]:
        return {
            'named': cls.named,
            'namespaced': cls.namespaced,
            'annotations': cls.annotations,
            'labels': cls.labels,
            'pluralname': cls.pluralname
        }


class ResourceMetaclass(pydantic.main.ModelMetaclass):

    def __new__(
        cls,
        name: str,
        bases: tuple[type[Any]],
        namespace: dict[str, Any],
        **params: Any
    ) -> 'ResourceMetaclass':
        if not namespace.pop('__abstract__', False):
            ResourceConfig: type[Any] | None = namespace.get('Config')
            annotations = namespace.setdefault('__annotations__', {})

            # To support dynamic subclassing, inspect the bases to determine if this
            # class directly inherits from a Resource subclass. Use an instance
            # check instead of a subclass check to remove the need for circular
            # imports. Reconstruct the subclass parameters from the existing class.
            for b in bases:
                if isinstance(b, ResourceMetaclass) and hasattr(b, '_version'):
                    group, version = api_version = b._version # type: ignore

                    # These are needed to initialize the superclass if this is a
                    # third-generation subclass of Resource.
                    params['group'] = group
                    params['version'] = version

                    # Get the Config class from parent.
                    ResourceConfig = getattr(b, 'Config', None)

                    break
            else:
                group = params.get('group')
                version = api_version = params.get('version')
                ResourceConfig = namespace.get('Config')
            if group is not None:
                api_version = f'{group}/{version}'

            # Create or update the config.
            if ResourceConfig is None:
                ResourceConfig = namespace['Config'] = type('Config', (Config,), {})
            assert ResourceConfig is not None
            attrs: dict[str, Any] = {
                'name': str.lower(name),
                **{k: v for k, v in Config.__dict__.items() if not k.startswith('_')},
                **{k: v for k, v in ResourceConfig.__dict__.items()
                if not k.startswith('_')}
            }
            attrs.setdefault('pluralname', engine.plural(attrs['name']))
            ResourceConfig = namespace['Config'] = type('Config', (Config,), attrs)

            # Create the `spec` field if the user has not specified it.
            if annotations.get('spec'):
                ResourceSpec: type[pydantic.BaseModel] = annotations['spec']
                assert isinstance(ResourceSpec, pydantic.main.ModelMetaclass)
                namespace.setdefault(
                    'spec',
                    pydantic.Field(
                        default=...,
                        title=ResourceSpec.__name__,
                        description=(
                            f'The `spec` field defines the specification of the '
                            f'desired behavior or state of the **{name}**.'
                        )
                    )
                )

            # Ensure that the apiVersion, kind and metadata fields are present. Defaults
            # and constraints are set during subclass initialization.
            Metadata = annotations.setdefault('metadata', ObjectMeta)
            annotations['metadata'] = Metadata.clone(**ResourceConfig.dict())
            annotations.update({
                'api_version': Literal[api_version], # type: ignore
                'kind': Literal[name], # type: ignore
            })
            namespace.update({
                '__annotations__': annotations,
                '_version': (group, version),
                'api_version': pydantic.Field(
                    alias='apiVersion',
                    description=(
                        'The `apiVersion` field defines the versioned schema of this '
                        'representation of an object. Servers should convert '
                        'recognized schemas to the latest internal value, and may '
                        'reject unrecognized values.'
                    )
                ),
                'kind': pydantic.Field(
                    description=(
                        'The `kind` field is a string value representing the REST resource this '
                        'object represents. Servers may infer this from the endpoint '
                        'the client submits requests to. Cannot be updated. '
                        'In `CamelCase`.'
                    )
                ),
                'metadata': pydantic.Field(
                    default=...,
                    title='Metadata',
                    description="Standard object's metadata."
                )
            })

            # Determine if there is a status model.
            ResourceStatus: Any | None = namespace.pop('Status', None)
            if ResourceStatus is not None:
                annotations.update({
                    'status': ResourceStatus.model | None
                })
                namespace.update({
                    'status': pydantic.Field(
                        default=None,
                        title='Status',
                        description=f'Most recently observed status of the `{name}`.'
                    )
                })

        new_cls = super().__new__(cls, name, bases, namespace, **params) # type: ignore
        return new_cls # type: ignore

    def __getitem__(
        self: T,
        k: EllipsisType | tuple[EllipsisType, bool]
    ) -> T:
        namespace: dict[str, Any] = {}
        config: type[Config] = getattr(self, 'Config', Config)
        return type(self.__name__, (self,), { # type: ignore
            **namespace,
            'Config': config,
            '__annotations__': {
                'metadata': BaseMeta.clone(
                    pristine=True,
                    **config.dict()
                )
            }
        })


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class ListMeta(pydantic.BaseModel):
    resource_version: str | None = pydantic.Field(
        default=None,
        alias='resourceVersion',
        title='Resource Version',
        description=(
            "String that identifies the server's internal version of this object "
            "that can be used by clients to determine when objects have changed. "
            "Value must be treated as opaque by clients and passed unmodified "
            "back to the server. Populated by the system. Read-only. "
        )
    )

    self_link: str | None = pydantic.Field(
        default=None,
        alias='selfLink',
        title='Self Link',
        description=(
            "SelfLink is a URL representing this object. Populated by the system. Read-only."
        )
    )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re

from .symbolicname import SymbolicName


pattern: str = r"^[\w'\-,.][^0-9_!¡?÷?¿/\\+=@#$%ˆ&*(){}|~<>;:[\]]{1,}$"


class RomanizedName(SymbolicName[2, 20, pattern]):
    """A string that contains a romanized personal name. A personal name is
    defined as follows:
    
    - Is at least 2 characters.
    - Begins with a word character, but not a number.
    - Contains at least one uppercase character for each space, dot
      or hyphen in the name, or consists of all uppercase characters.

    Automatic sanitation and transformation is performed on:

    - Stripping non-word characters from the beginning and the end of the
      name.
    - Normalizing whitespace, dots or hyphens to one character.
    - If the name does not contain a space, dot or hyphen, then the first
      character is capitalized.
    """

    @classmethod
    def validate(cls, v: str) -> str:
        v = re.sub(r'[^\w]+$', '', v)
        v = re.sub(r'^[^\w]+', '', v)
        v = re.sub(r'\s+', ' ', v)
        v = re.sub(r'\.+', '.', v)
        v = re.sub(r'\-', '-', v)
        if str.isnumeric(v[0]):
            raise ValueError("Name cannot start with a number.")

        length: int = len(v)
        characters = list(v)
        for i, char in enumerate(characters):
            if (i + 1) == length:
                continue

            # If the character is a hyphen, then the next character
            # is capitalized.
            if char == '-':
                characters[i + 1] = str.upper(v[i + 1])

        v = ''.join(characters)
        if str.upper(v) != v:
            # If the name does not contain any spaces, hyphens or dots,
            # then it must start with a capital.
            if not re.findall(r'(\s|\-|\.)', v):
                characters[0] = str.upper(characters[0])

        return super().validate(str.join('', characters))

    def __repr__(self) -> str: # pragma: no cover
        return f'RomanizedName({self})'